/* cbs.cpp
 *
 * General routines for cubic b-splines.
 * 
 * History
 *  2011/06/24: First version
 *
 * Note
 *  CBS is partly based on the original source code of BBS.
 *
 * (c)2011, Florent Brunet.
 */

/* This file is part of CBS.
 *
 * CBS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CBS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CBS.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "cbs.h"
#include <stdlib.h>
#include <math.h>

#define max(a,b) (((a)>=(b)) ? (a) : (b))
#define min(a,b) (((a)<=(b)) ? (a) : (b))

/* Normalize the values in "t". "nb_t" is the size of the "t" and "nt" arrays. */
void normalize(double tmin, double tmax, int npts, double *t, int nb_t, double* nt) {
    int ninter = npts - 3;
    double width_inter = (tmax - tmin) / ninter;
    
    for (int i = 0 ; i < nb_t; ++i) {
        if (t[i] == tmax)
            nt[i] = 1.0;
        else if (t[i] < tmin)
            nt[i] = (t[i] - tmin) / width_inter;
        else if (t[i] > tmax)
            nt[i] = (t[i] - tmin) / width_inter - ninter;
        else {
            double scaled = (t[i] - tmin) / width_inter;
            nt[i] = scaled - floor(scaled);
        }
    }
}


/* Same as "normalize" but computes also the interval number to which the t's belongs. */
void normalize_with_inter(double tmin, double tmax, int npts, double *t, int nb_t, double* nt, int *inter) {
    int ninter = npts - 3;
    double width_inter = (tmax - tmin) / ninter;
    
    for (int i = 0 ; i < nb_t; ++i) {
        if (t[i] == tmax) {
            nt[i] = 1.0;
            inter[i] = ninter - 1;
        } else if (t[i] < tmin) {
            nt[i] = (t[i] - tmin) / width_inter;
            inter[i] = -1;
        } else if (t[i] > tmax) {
            nt[i] = (t[i] - tmin) / width_inter - ninter;
            inter[i] = ninter;
        } else {
            double scaled = (t[i] - tmin) / width_inter;
            inter[i] = (int)floor(scaled);
            nt[i] = scaled - inter[i];
        }
    }
}




/* Evaluate the 4 basis at nt. "nt" must be a normalized value. */
void eval_basis(double nt, double *val_basis) {
    double nt2 = nt * nt;
    double nt3 = nt2 * nt;
    val_basis[0] = (-nt3 + 3.0*nt2 - 3.0*nt + 1.0) / 6.0;
    val_basis[1] = (3.0*nt3 - 6.0*nt2 + 4.0) / 6.0;
    val_basis[2] = (-3.0*nt3 + 3.0*nt2 + 3.0*nt + 1.0) / 6.0;
    val_basis[3] = nt3 / 6.0;
}

/* Evaluate the 4 basis first derivatives at nt. "nt" must be a normalized value. */
void eval_basis_d(double nt, double *val_basis) {
    double nt2 = nt * nt;
    val_basis[0] = (-nt2 + 2*nt - 1) / 2.0;
    val_basis[1] = (3.0*nt2 - 4.0*nt) / 2.0;
    val_basis[2] = (-3*nt2 + 2*nt + 1) / 2.0;
    val_basis[3] = nt2 / 2.0;
}

/* Evaluate the 4 basis second derivatives at nt. "nt" must be a normalized value. */
void eval_basis_dd(double nt, double *val_basis) {
    val_basis[0] = -nt + 1.0;
    val_basis[1] = 3.0*nt - 2.0;
    val_basis[2] = -3.0*nt + 1.0;
    val_basis[3] = nt;
}



/* Pointer type for a basis function evaluation */
typedef void (*basis_func_t)(double,double*);

/* Return a pointer to the basis function that correspond to the "order" of derivation. */
basis_func_t get_basis_ptr(int order) {
    switch (order) {
        case 0: return eval_basis;
        case 1: return eval_basis_d;
        case 2: return eval_basis_dd;
    }
    return NULL;
}


double get_deriv_fact(cbs_t* cbs, int order) {
    double s = (cbs->tmax - cbs->tmin) / (cbs->npts - 3);
    
    return 1.0 / pow(s,order);
}



/* Evaluate a spline (or its derivatives, according to the basis functions utilized).
 * - cbs: bicubic bspline structure.
 * - ctrlpts: control points (valdim x npts matrix).
 * - t: locations where the b-spline is evaluated.
 * - nval: number of values in t.
 * - val: valdim x npts matrix containing the result.
 * - dt: order of derivation. */
void eval(cbs_t *cbs, double *ctrlpts, double *t, int nval, double *val, int dt) {
    double *nt = (double*)malloc(nval * sizeof(double));
    int *inter = (int*)malloc(nval * sizeof(int));

    // Compute the normalized evaluation values and their interval numbers
    normalize_with_inter(cbs->tmin, cbs->tmax, cbs->npts, t, nval, nt, inter);
    
    for (int k = 0; k < nval; ++k) {
        double basis[4];
        basis_func_t b_func = get_basis_ptr(dt);
        double fact = get_deriv_fact(cbs, dt);

        b_func(nt[k], basis);
        
        for (int d = 0; d < cbs->valdim; ++d)
            val[cbs->valdim*k+d] = 0.0;
 
        for (int i = 0; i < 4; ++i) {
            double bas = basis[i];
            int ind = cbs->valdim * (i+inter[k]);
            for (int d = 0; d < cbs->valdim; ++d)
                val[cbs->valdim*k+d] += ctrlpts[ind++] * bas;
        }
        
        for (int d = 0; d < cbs->valdim; ++d)
            val[cbs->valdim*k+d] *= fact;
    }
    
    free(nt);
    free(inter);
}



/* Build a sparse colocation matrix.
 * This function does not handle sites that are outside of the domain (an error is issued in such cases).
 * ARGUMENTS:
 * - cbs: b-spline structure.
 * - t: sites used to build the colocation matrix.
 * - nsites: number of sites (number of elements in t).
 * - pr: array containing the values of the matrix (size: number of non-zero elements, i.e. 16*nsites).
 * - ir: ir(i) is the row index (0-based) of pr(i). The sizes of ir and pr are equal.
 * - jc: jc(j) is the shift in pr (and ir) for the j-th column. numel(jc) = 1+nptsx*nptsy (number of columns). 
 * RETURN VALUE:
 * This function return 0 if successful, an error code otherwise:
 * 0: successful.
 * 1: a colocation site was ouside of the spline definition domain. */
int coloc(cbs_t *cbs, double *t, int nsites, double *pr, mwIndex *ir, mwIndex *jc) {
    int ret_code = 0;
    double *nt = (double*)malloc(nsites * sizeof(double));
    int *intert = (int*)malloc(nsites * sizeof(int));
    double basis_t[4];

    int npts = cbs->npts;
    
    mwIndex *nb_elem_col = (mwIndex*)calloc(npts, sizeof(mwIndex));
    mwIndex *cur_ind = (mwIndex*)malloc(npts*sizeof(mwIndex));
    
    // Compute the normalized evaluation values and their interval numbers
    normalize_with_inter(cbs->tmin, cbs->tmax, cbs->npts, t, nsites, nt, intert);
    
    // FILLING jc
    // First, compute the number of elements in each column
    // Check at the same time if the sites are inside the definition domain.
    for (int k = 0; k < nsites; ++k) {
        int I = intert[k];
        
        if (I < 0 || I > cbs->npts-4) {
            ret_code = 1;
            goto error;
        }
        
        for (int i = 0; i < 4; ++i)
            nb_elem_col[i+I] += 1;
    }
    
    // Second, compute jc from nb_elem_col
    for (int k = 1; k <= npts; ++k)
        jc[k] = jc[k-1] + nb_elem_col[k-1];
    
    // FILLING pr and ir
    // cur_ind contains the current shift in pr (and ir) for each column
    for (int k = 0; k < npts; ++k)
        cur_ind[k] = jc[k];
    
    for (int k = 0; k < nsites; ++k) {
        eval_basis(nt[k], basis_t);

        for (int i = 0; i < 4; ++i) {
            int col = i + intert[k];
            pr[cur_ind[col]] = basis_t[i];
            ir[cur_ind[col]] = k;
            cur_ind[col] += 1;
        }
    }

error:
    
    free(nt);
    free(intert);
    free(nb_elem_col);
    free(cur_ind);
    
    return ret_code;
}


/* Build a sparse colocation matrix that accounts for derivatives.
 * This function does not handle sites that are outside of the domain (an error is issued in such cases).
 * ARGUMENTS:
 * - cbs: b-spline structure.
 * - t: sites used to build the colocation matrix.
 * - nsites: number of sites (number of elements in t).
 * - dt: order of derivation.
 * - pr: array containing the values of the matrix (size: number of non-zero elements, i.e. 4*nsites).
 * - ir: ir(i) is the row index (0-based) of pr(i). The sizes of ir and pr are equal.
 * - jc: jc(j) is the shift in pr (and ir) for the j-th column. numel(jc) = 1+npts (number of columns). 
 * RETURN VALUE:
 * This function return 0 if successful, an error code otherwise:
 * 0: successful.
 * 1: a colocation site was ouside of the spline definition domain. */
int coloc_deriv(cbs_t *cbs, double *t, int nsites, int dt, double *pr, mwIndex *ir, mwIndex *jc) {
    int ret_code = 0;
    double *nt = (double*)malloc(nsites * sizeof(double));
    int *intert = (int*)malloc(nsites * sizeof(int));
    basis_func_t b_func = get_basis_ptr(dt);
    double fact = get_deriv_fact(cbs, dt);
    double basis_t[4];

    int npts = cbs->npts;
    
    mwIndex *nb_elem_col = (mwIndex*)calloc(npts, sizeof(mwIndex));
    mwIndex *cur_ind = (mwIndex*)malloc(npts*sizeof(mwIndex));
    
    // Compute the normalized evaluation values and their interval numbers
    normalize_with_inter(cbs->tmin, cbs->tmax, cbs->npts, t, nsites, nt, intert);
    
    // FILLING jc
    // First, compute the number of elements in each column
    // Check at the same time if the sites are inside the definition domain.
    for (int k = 0; k < nsites; ++k) {
        int I = intert[k];
        
        if (I < 0 || I > cbs->npts-4) {
            ret_code = 1;
            goto error;
        }
        
        for (int i = 0; i < 4; ++i)
            nb_elem_col[i+I] += 1;
    }
    
    // Second, compute jc from nb_elem_col
    for (int k = 1; k <= npts; ++k)
        jc[k] = jc[k-1] + nb_elem_col[k-1];
    
    // FILLING pr and ir
    // cur_ind contains the current shift in pr (and ir) for each column
    for (int k = 0; k < npts; ++k)
        cur_ind[k] = jc[k];
    
    for (int k = 0; k < nsites; ++k) {
        b_func(nt[k], basis_t);

        for (int i = 0; i < 4; ++i) {
            int col = i + intert[k];
            pr[cur_ind[col]] = fact*basis_t[i];
            ir[cur_ind[col]] = k;
            cur_ind[col] += 1;
        }
    }

error:
    
    free(nt);
    free(intert);
    free(nb_elem_col);
    free(cur_ind);
    
    return ret_code;
}
